/*---------------------------------------------------------------------------
GIF.CPP - Part of FAST Viewer project
Description: This file handles gif format. Although there are several
functions defined in this file, at present, only two are accessible from
outside: one displays gif files on screen and other reads information from a
gif image. All other functions are for internal use. This module may be used
in some other project as it is an independent one. Just include "gif.h" in
that project.
-----------------------------------------------------------------------------
Accessible functions:
int show_gif (char *name);
void read_gif_info (FILE *dataFile, gifInfoType &gifInfo);
---------------------------------------------------------------------------*/


#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <stdio.h>
#include "mode13h.h"


/*---------------------------------------------------------------------------
GLOBAL VARIABLES FOR GIF.CPP
---------------------------------------------------------------------------*/
FILE *dataFile;			// our gif file made global
unsigned long pwr2[17];		// to store different powers of 2 for speed
int nBitsIn;			// number of bits taken in from a char
int num;
int aChar;                      // a simple char read from the image
int blockLength;
long paletteOffset, dataOffset;	// positions where they begin in file


/*---------------------------------------------------------------------------
GIF INFORMATION HEADER
---------------------------------------------------------------------------*/
struct gifInfoType		// our home made gif information header
   {
   char version[7];		// version either GIF89a or GIF87a
   byte gcTable;		// global color table exists? answer: 0 or 1
   byte bitsPerPixel;		// bits per pixel --> global
   byte background;		// usually 0
   unsigned int xStart, yStart; // usually 0, 0
   int width, height;		// local width and height of the image
   byte lcTable;		// local color table exists
   byte interlaced;		// interlaced image?
   };



/*---------------------------------------------------------------------------
FUNCTION:	Initialize variables (called by main show_gif)
INPUT:		None
PROCESS:	Sets values for different global variables
OUTPUT:		None
---------------------------------------------------------------------------*/
void initialize (void)
   {
   int i;		// loop control variable
   nBitsIn = 8;		// set number of bits taken in to 8
   num = 0;
   pwr2[0] = 1;		// generate powers of 2
   for (i=1; i<17; i++) pwr2[i] = pwr2[i-1]*2;
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Get one bit from dataFile (called by read_code)
INPUT:		None
PROCESS:	Extracts one bit after reading a character from the file
		and remembers how many bits from current character have been
		extracted
OUTPUT:		Returns the bit read - either 0 or 1
---------------------------------------------------------------------------*/
int get_bit (void)
   {
   nBitsIn++;                        	// increment number of bits taken in
   if (nBitsIn == 9)			// a whole char done?
      {
      aChar = fgetc (dataFile);		// read another char and
      nBitsIn = 1;			// set number of bits taken in to 1
      num++;
      if (num == blockLength)
	 {
	 blockLength = aChar + 1;
	 aChar = fgetc (dataFile);
	 num = 1;
	 }
      }
   if ((aChar & pwr2[nBitsIn-1]) == 0) return 0;	// bit read 0 or 1?
   else return 1;
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Reads code from the dataFile (called by main show_gif)
INPUT:		Current size of code
PROCESS:	Calls get_bit many times to generate a code of given size
OUTPUT:		Required code
---------------------------------------------------------------------------*/
int read_code (int codeSize)
   {
   int i, code=0;			// initialize code to 0
   for (i=0; i<codeSize; i++)		// take bits and add them to code
      code += get_bit() * pwr2[i];      // after converting to decimal
   return code;				// return the read code
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Reads gif info
INPUT:		Name of file and header reference variable
PROCESS:	Opens up the file and reads in its header using several calls
		of fgetc()
OUTPUT:		Error if invalid header
---------------------------------------------------------------------------*/
int read_gif_info (FILE *dataFile, gifInfoType &gifInfo)
   {
   int i;		// lcv
   char abyte;		// temporary char variable
   initialize();	// initialize global variables

   for (i=0; i<6; i++) gifInfo.version[i] = fgetc (dataFile);
   gifInfo.version[6] = '\0';			// read version six bytes
   for (i=0; i<4; i++) (void) fgetc (dataFile);	// skip width and height (int)
   abyte = fgetc (dataFile);			// store next byte
   gifInfo.gcTable = ((abyte & 128) == 128);	// gcTable?
   gifInfo.bitsPerPixel = (abyte & 7) + 1;	// Bits per pixel
   gifInfo.background = fgetc (dataFile);	// read background
   abyte = fgetc (dataFile);			// next byte always 0
   if (abyte!=0) return 1;			// not 0? invalid gif image
   fgetpos (dataFile, &paletteOffset);		// skip palette for some time

   // read local data
   fseek (dataFile, pwr2[gifInfo.bitsPerPixel]*3, SEEK_CUR);
   abyte = fgetc (dataFile);			// skip next byte
   fread (&gifInfo.xStart, 2, 1, dataFile);	// read local upper coords
   fread (&gifInfo.yStart, 2, 1, dataFile);
   fread (&gifInfo.width,  2, 1, dataFile);	// read local width and height
   fread (&gifInfo.height, 2, 1, dataFile);
   abyte = fgetc (dataFile);			// store next byte
   gifInfo.lcTable = (((abyte & 128) == 128 ) ? 1 : 0);	// local color table?
   gifInfo.interlaced = (((abyte & 64) == 64) ? 1 : 0);	// interlaced image?
   fgetpos (dataFile, &dataOffset);		// store data for some time

   return 0;		// successful
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Decompresses and shows a gif image on screen
INPUT:		Name of the gif file
PROCESS:	Calls read_gif_info to read image information,
		checks for support,
		calls read_code to read codes of varying length,
		decompresses the image and shows it on the screen
OUTPUT:		Return error code if image can't be shown
---------------------------------------------------------------------------*/
int show_gif (char *name)
   {
   gifInfoType gifInfo;                 // info about current gif file
   byte abyte;				// byte buffer
   byte r, g, b;			// for palette
   int  i;				// for loops, etc.
   int  codeSize, code, inCode, 	// for LZW decompression
	curCode, oldCode,
	prefix[4096], suffix[4096];
   int  clearCode, eofCode, freeCode,	// for GIF variation of LZW
	firstFree, maxCode,
	initCodeSize, bitMask, finChar,
	outCount, outCode[1025];
   int  x, y;                           // position of pixel to plot


   dataFile = fopen (name, "rb");       // open file
   if (dataFile == NULL) return 6;	// could not open?
   if (read_gif_info (dataFile, gifInfo) != 0)
      { fclose (dataFile); return 5; }	// invalid gif image
   if (strcmpi (gifInfo.version, "GIF89a") == 0)
      { fclose (dataFile); return 1; }	// unsupported gif version
   else if (strcmpi (gifInfo.version, "GIF87a") != 0)
      { fclose (dataFile); return 5; }	// invalid header
   if (gifInfo.gcTable == 0)
      { fclose (dataFile); return 2; }	// can't do without global color table
   if (gifInfo.lcTable != 0)            // can't handle local color table
      { fclose (dataFile); return 3; }
   if (gifInfo.interlaced != 0)		// can't handle interlaced image
      { fclose (dataFile); return 4; }


   // read palette
   set_mcga();
   fseek (dataFile, paletteOffset, SEEK_SET);	// go to where palette begins
   for (i=0; i<pwr2[gifInfo.bitsPerPixel]; i++)	// start reading it
      {
      r = fgetc (dataFile); r = r>>2;
      g = fgetc (dataFile); g = g>>2;
      b = fgetc (dataFile); b = b>>2;
      set_palette (i, r, g, b);
      }


   // read data
   fseek (dataFile, dataOffset, SEEK_SET);	// go to where image begins
   codeSize = fgetc (dataFile);			// read code size
   clearCode = pwr2[codeSize];			// clear code = 2^codeSize
   eofCode = clearCode + 1;			// end of file code
   freeCode = firstFree = clearCode + 2;	// free entry code
   codeSize++;					// actual code size
   initCodeSize = codeSize;			// which is initial too
   maxCode = pwr2[codeSize];
   bitMask = pwr2[gifInfo.bitsPerPixel]-1;	// for extracting bits

   blockLength = fgetc (dataFile) + 1;
   outCount = 0;				// number of pixels stacked
   x = gifInfo.xStart;				// x and y coordinates for a
   y = gifInfo.yStart;				// pixel


   // start decoding and show image
   do
      {
      code = read_code (codeSize);		// read a code of codeSize
      if (code == eofCode) break;		// end of file code?
      if (code == clearCode)			// clear code?
	 {
	 codeSize = initCodeSize;		// set to initial code size
	 maxCode = pwr2[codeSize];
	 freeCode = firstFree;    		// set to initial free code
	 code = read_code (codeSize);		// read next code
	 oldCode = curCode = code;		// make it current and old
	 finChar = code & bitMask;		// extract bits to plot
	 put_pixel (x++, y, finChar);		// plot it
	 if (x >= gifInfo.width) { x=0; y++; }	// next line?
	 }
      else					// not clear code
	 {
	 inCode = curCode = code;		// current code and in code

	 if (code >= freeCode)			// not in the table?
	    {
	    curCode = oldCode;
	    outCode [outCount++] = finChar;
	    }

	 if (curCode > bitMask) do
	    {
	    outCode [outCount++] = suffix [curCode];
	    curCode = prefix [curCode];
	    } while (curCode > bitMask);

	 finChar = curCode & bitMask;		// extract bits to plot
	 outCode [outCount++] = finChar;	// stack it

	 for (i=outCount-1; i>=0; i--)		// plot the stacked pixels
	    {
	    put_pixel (x++, y, outCode[i]);
	    if (x >= gifInfo.width) { x=0; y++; }
	    }

	 outCount = 0;				// empty stack
	 prefix [freeCode] = oldCode;
	 suffix [freeCode++] = finChar;
	 oldCode = inCode;
	 if (freeCode >= maxCode && codeSize < 12)
	    { codeSize++; maxCode *= 2; }
	 }
      } while (code != eofCode);	// end of file code? exit

   fclose (dataFile);			// close the data file
   getch();				// wait for a key press
   set_text();				// return to text mode
   return 0;				// reached here? then no error
   }
/*-------------------------------------------------------------------------*/